@@ -1,12 +1,17 @@
 # -*- coding: utf-8 -*-
 
-
-from configparser import SafeConfigParser, _UNSET
+from configparser import SafeConfigParser
 import inspect
 import logging
 import os
 import pickle  # nosec
 import pkg_resources
+from typing import (
+    Optional,
+    List,
+    Union,
+    Text
+)
 
 from enhancements.modules import get_module_class, Module
 
@@ -17,14 +22,21 @@ class DefaultConfigNotFound(Exception):
 
 class ExtendedConfigParser(SafeConfigParser):
 
-    def __init__(self, productionini=None, defaultini='default.ini', package=None, env_name='ENHANCED_CONFIG_FILE', modules_from_file=False):
+    def __init__(
+        self,
+        productionini: Optional[Text] = None,
+        defaultini: Text = 'default.ini',
+        package: Optional[Text] = None,
+        env_name: Text = 'ENHANCED_CONFIG_FILE',
+        modules_from_file: bool = False
+    ):
         super(ExtendedConfigParser, self).__init__(allow_no_value=True)
-        self.defaultini = defaultini
-        self.package = package
-        self.default_config = self._get_default_config()
-        self.production_config = None
-        self.configfiles = []
-        self.modules_from_file = modules_from_file
+        self.defaultini: Text = defaultini
+        self.package: Optional[Text] = package
+        self.default_config: Text = self._get_default_config()
+        self.production_config: Optional[Text] = None
+        self.configfiles: List[Text] = []
+        self.modules_from_file: bool = modules_from_file
 
         self._read_default_config()
 
@@ -39,7 +51,7 @@ class ExtendedConfigParser(SafeConfigParser):
         if self.production_config:
             self.append(self.production_config)
 
-    def _get_default_config(self):
+    def _get_default_config(self) -> Text:
         packages = []
         if self.package:
             packages.append(self.package)
@@ -55,7 +67,7 @@ class ExtendedConfigParser(SafeConfigParser):
 
         raise DefaultConfigNotFound()
 
-    def _read_default_config(self):
+    def _read_default_config(self) -> None:
         if self.default_config:
             logging.debug("Using default config: %s", self.default_config)
             self.append(self.default_config)
@@ -66,12 +78,12 @@ class ExtendedConfigParser(SafeConfigParser):
         except Exception:
             logging.exception("error reading %s", configpath)
 
-    def copy(self):
+    def copy(self) -> 'ExtendedConfigParser':
         """ create a copy of the current config
         """
         return pickle.loads(pickle.dumps(self))  # nosec
 
-    def append(self, configpath):
+    def append(self, configpath: Text) -> None:
         self.configfiles.append(configpath)
         if not configpath:
             return
@@ -84,10 +96,10 @@ class ExtendedConfigParser(SafeConfigParser):
                 configpath
             )
 
-    def getlist(self, section, option, sep=',', chars=None):
+    def getlist(self, section: Text, option: Text, sep: Text = ',', chars: Optional[Text] = None) -> List[Text]:
         return [chunk.strip(chars) for chunk in self.get(section, option).split(sep)]
 
-    def _getmodule_option(self, section, option):
+    def _getmodule_option(self, section: Text, option: Text) -> Module:
         """ get a module class from config file
         """
         module = self.get(section, option)
@@ -102,7 +114,7 @@ class ExtendedConfigParser(SafeConfigParser):
             )
         return values[0]
 
-    def _getmodule_section(self, section):
+    def _getmodule_section(self, section: Text) -> Optional[Module]:
         if not self.has_section(section):
             raise ValueError('Config section does not exist! Module not loaded.')
         if not self.has_option(section, 'enabled') or not self.has_option(section, 'class'):
@@ -111,26 +123,29 @@ class ExtendedConfigParser(SafeConfigParser):
             return None
         return self.getmodule(section, 'class')
 
-    def getmodule(self, section, option=None):
+    def getmodule(self, section: Text, option: Optional[Text] = None) -> Optional[Module]:
         if option:
             return self._getmodule_option(section, option)
         return self._getmodule_section(section)
 
-    def getplugins(self, module_prefix):
+    def getplugins(self, module_prefix) -> List[Module]:
         plugins = []
-        if isinstance(module_prefix, (str, Module)) or issubclass(module_prefix, Module):
-            module_prefix = module_prefix.CONFIG_PREFIX
-        else:
+        if isinstance(module_prefix, Module) or issubclass(module_prefix, Module):
+            if module_prefix.CONFIG_PREFIX:
+                module_prefix = module_prefix.CONFIG_PREFIX
+        elif not isinstance(module_prefix, str):
             raise ValueError("Not a valid module prefix. Only strings and module are supported.")
 
         for section in self.sections():
             if section.startswith('{}:'.format(module_prefix)):
                 if self.getboolean(section, 'enabled'):
-                    plugins.append(self.getmodule(section, 'class'))
+                    module = self.getmodule(section, 'class')
+                    if module:
+                        plugins.append(module)
         return plugins
 
-    def getboolean_or_string(self, section, option, *, raw=False, vars=None, fallback=_UNSET):
+    def getboolean_or_string(self, section: Text, option: Text) -> Union[bool, Text]:
         try:
-            return self.getboolean(section, option, raw=raw, vars=vars, fallback=fallback)
+            return self.getboolean(section, option)
         except ValueError:
-            return self.get(section, option, raw=raw, vars=vars, fallback=fallback)
+            return self.get(section, option)
